from configobj import ConfigObj


def dict_to_ini_file(dict_obj, file_path, sort_keys=True):
    config = ConfigObj(indent_type='    ', write_empty_values=True)
    config.filename = file_path
    _first_line = True
    if sort_keys:
        # Sort by connection name in order to make the credential file more readable
        dict_obj = dict(sorted(dict_obj.items()))
    for k, v in dict_obj.items():
        config[k] = v

        if _first_line:
            _first_line = False
        else:
            # Hack for adding a blank line between sections: Insert an empty comment before each section
            # https://sourceforge.net/p/configobj/mailman/message/24432354/
            config.comments[k].insert(0, '')
    config.write()



import lhub
from .connection_manager import LogicHubConnection
from .actions import Actions


class LogicHubCLI:

    def __init__(self, instance_name, log_level=None):
        # ToDo:
        #  * Move the logging function out of lhub and into lhub_cli
        #  * Standardize better w/ the "logging" package
        self.instance_name = instance_name
        self.__config = LogicHubConnection(instance_name)
        self.session = lhub.LogicHub(
            **self.__config.credentials.to_dict(),
            api_key=self.__config.credentials.api_key,
            password=self.__config.credentials.password,
            log_level=log_level
        )
        self.log = self.session.api.log
        self.actions = Actions(
            session=self.session,
            config=self.__config,
            instance_label=self.__config.instance
        )



from pathlib import Path
import os
from configobj import ConfigObj
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from lhub import LogicHub
from .encryption import Encryption
import getpass
from .common.config import dict_to_ini_file
from .common.shell import query_yes_no
from requests.exceptions import SSLError

# Over-writable/configurable static variables
LHUB_CONFIG_PATH = os.path.join(str(Path.home()), ".logichub")
CREDENTIALS_FILE_NAME = "credentials"
PREFERENCES_FILE_NAME = "preferences"

# Ensure that the path exists
__lhub_path = Path(LHUB_CONFIG_PATH)
__lhub_path.mkdir(parents=True, exist_ok=True)


@dataclass_json
@dataclass
class _PreferenceMain:
    default_instance: str = None


@dataclass_json
@dataclass
class _PreferenceCommands:
    table_style: str = None


@dataclass_json
@dataclass
class Preferences:
    main: _PreferenceMain = None
    commands: _PreferenceCommands = None

    def __post_init__(self):
        self.preferences_file_name = PREFERENCES_FILE_NAME
        self.preferences_path = os.path.join(LHUB_CONFIG_PATH, self.preferences_file_name)
        self.get_preferences()

    def save_preferences_file(self):
        dict_to_ini_file(self.to_dict(), self.preferences_path)

    def get_preferences(self):
        if not os.path.exists(self.preferences_path):
            self.main = _PreferenceMain()
            self.commands = _PreferenceCommands()
            self.save_preferences_file()
            return

        preferences_obj = ConfigObj(self.preferences_path)
        if not self.main:
            self.main = _PreferenceMain(**{
                k: v for k, v in preferences_obj.dict().get("main", {}).items()
            })
        if not self.commands:
            self.commands = _PreferenceCommands(**{
                k: v for k, v in preferences_obj.dict().get("commands", {}).items()
            })


@dataclass_json
@dataclass
class Connection:
    # Deliberately left out password and api_token so that those sensitive fields do not show in JSON form
    hostname: str
    username: str
    verify_ssl: bool

    def __init__(self, name, **kwargs):
        self.connection_name = name
        self.hostname = kwargs.get("hostname")
        self.api_key = kwargs.get("api_key")
        self.username = kwargs.get("username")
        self.password = kwargs.get("password")
        self.verify_ssl = kwargs.get("verify_ssl") or True
        if isinstance(self.verify_ssl, str):
            self.verify_ssl = not self.verify_ssl.lower() == 'false'
        assert self.hostname, "Server hostname not provided"
        if not self.api_key:
            assert self.password, "Neither an API key nor a password were provided in the connection config"
            assert self.username, "Username not provided"

# https://betterprogramming.pub/advanced-python-9-best-practices-to-apply-when-you-define-classes-871a27af658b#:~:text=Define%20Meaningful%20String%20Representations
    def __repr__(self):
        return self.connection_name

    def __str__(self):
        return self.connection_name



class LhubConfig:
    __full_config: ConfigObj = None
    credentials_file_name = CREDENTIALS_FILE_NAME

    def __init__(self):
        self.credentials_path = os.path.join(LHUB_CONFIG_PATH, self.credentials_file_name)
        self.__load_credentials_file()
        self.encryption = Encryption(LHUB_CONFIG_PATH)

    def __load_credentials_file(self):
        if not os.path.exists(self.credentials_path):
            dict_to_ini_file({}, self.credentials_path)
        self.__full_config = ConfigObj(self.credentials_path)

    def reload(self):
        self.__load_credentials_file()

    def update_connection(self, instance_label, **kwargs):
        if not self.__full_config.get(instance_label):
            self.__full_config[instance_label] = kwargs
        else:
            self.__full_config[instance_label].update(kwargs)
        dict_to_ini_file(self.__full_config.dict(), self.credentials_path)

    def delete_connection(self, instance_label):
        if not self.__full_config.get(instance_label):
            return
        else:
            del self.__full_config[instance_label]
        dict_to_ini_file(self.__full_config.dict(), self.credentials_path)

    def get_instance(self, instance_label):
        if instance_label not in self.__full_config:
            return
        # Make a copy of the dict, otherwise this will only work once, and it
        # will fail with a decryption error any subsequent calls for the same instance
        _credentials = {k: v for k, v in self.__full_config[instance_label].items()}
        for k in _credentials:
            if k in ('password', 'api_key'):
                _credentials[k] = self.encryption.decrypt_string(_credentials[k])
        return Connection(name=instance_label, **_credentials)

    def create_instance(self, instance_label, server=None, auth_type=None, api_key=None, username=None, password=None, verify_ssl=None):
        instance_label = instance_label.strip()
        if instance_label in self.__full_config:
            raise ValueError(f"An instance already exists by the name {instance_label}")
        server = server.strip() if server else None
        auth_type = auth_type.strip() if auth_type else None
        api_key = api_key.strip() if api_key else None
        username = username.strip() if username else None
        password = password.strip() if password else None
        verify_ssl = verify_ssl if isinstance(verify_ssl, bool) else None

        if password and not api_key:
            # If a password was provided but no API key, assume password auth
            auth_type = 'password'
        if api_key and not password:
            # If a API key was provided but no password, assume password auth
            auth_type = 'api_key'

        while not server:
            server = input("Server hostname or IP: ").strip()

        while auth_type not in ('password', 'api_key'):
            auth_type = input('Enter "1" for password auth or "2" for API token auth:\n').strip()
            if auth_type == '1':
                auth_type = "password"
            elif auth_type == '2':
                auth_type = 'api_key'
            else:
                print('Invalid input\n')

        # Password auth
        if auth_type == 'password':
            api_key = None
            while not username:
                username = input("Username: ")
            while not password:
                password = getpass.getpass()

        # API token auth
        else:
            username = password = None
            while not api_key:
                api_key = getpass.getpass(prompt='API Token: ')

        # verify_ssl = query_yes_no('Verify SSL?', 'y')
        if verify_ssl is None:
            verify_ssl = query_yes_no('Verify SSL?', 'y')

        try:
            _ = LogicHub(hostname=server, username=username, password=password, api_key=api_key, verify_ssl=verify_ssl)
        except SSLError as err:
            verify_ssl = not query_yes_no('SSL certificate verification failed. Disable SSL verification?')
            if verify_ssl:
                raise err
            else:
                _ = LogicHub(hostname=server, username=username, password=password, api_key=api_key, verify_ssl=verify_ssl)

        if password:
            password = self.encryption.encrypt_string(password)
            self.update_connection(instance_label, hostname=server, username=username, password=password, verify_ssl=verify_ssl)
        else:
            api_key = self.encryption.encrypt_string(api_key)
            self.update_connection(instance_label, hostname=server, api_key=api_key, verify_ssl=verify_ssl)
        self.reload()

    def list_configured_instances(self):
        return sorted(self.__full_config.dict().keys())

    def print_configured_instances(self):
        instances = self.list_configured_instances()
        if not instances:
            print("No stored instances")
        else:
            print(f"Stored instances:")
            for i in instances:
                print(f"    {i}")


class LogicHubConnection:
    __instance = None
    # preferences: Preferences = None
    credentials = None

    # ToDo Hide this (dunder) when finished developing
    config: LhubConfig = None

    def __init__(self, instance_alias=None):
        self.config = LhubConfig()
        # ToDo Not yet used, but the groundwork has been laid. Revisit and enable this when ready to begin putting it to use.
        # if not self.preferences:
        #     self.preferences = Preferences()
        if instance_alias:
            self.instance = instance_alias

    @property
    def all_instances(self):
        return self.config.list_configured_instances()

    @property
    def instance(self):
        if not self.__instance:
            raise ValueError("Instance not set")
        return self.__instance

    @instance.setter
    def instance(self, name: str):
        name = name.strip()
        self.config.reload()
        _new_credentials = self.config.get_instance(name)
        if not _new_credentials:
            print(f"No instance found by name \"{name}.\" Creating new connection...")
            self.config.create_instance(name)
            _new_credentials = self.config.get_instance(name)

        self.__instance = name
        self.credentials = _new_credentials

        # ToDo Logic here for prompting for credentials, test them, and then save an updated credentials file

        # ToDo Also use this class to invoke the LogicHub class from lhub_pw.py. When the instance is changed, create a new LogicHub instance



import cmd
from .connection_manager import LogicHubConnection
from lhub.log import Logger
from colorama import init as init_colorama
from colorama import Fore, Back, Style

log = Logger()

# ToDo This didn't work either.
init_colorama()


# ToDo make sure Shell.session doesn't have to be set explicitly in order to use this
class Shell(cmd.Cmd):
    # Override the default prompt of "(cmd) "
    # prompt = "(logichub) "
    # Fancy colorful version
    # prompt = '\x1b[6;30;47m' + 'Logic' + '\x1b[6;31;47m' + 'Hub' + '\x1b[0m % '
    prompt = Fore.BLACK + Back.WHITE + 'Logic' + Fore.RED + 'Hub' + Style.RESET_ALL + ' % '
    session: LogicHubConnection = None

    def __init__(self):
        super().__init__()
        if not self.session:
            self.session = LogicHubConnection()

    # ToDo Finish or remove; overriding in order to try to allow keyboard interrupts
    def cmdloop(self, intro=None):
        """
        Repeatedly issue a prompt, accept input, parse an initial prefix off the
        received input, and dispatch to action methods, passing them the
        remainder of the line as argument.

        :param intro:
        :return:
        """

        self.preloop()
        if self.use_rawinput and self.completekey:
            try:
                import readline
                self.old_completer = readline.get_completer()
                readline.set_completer(self.complete)
                readline.parse_and_bind(self.completekey+": complete")
            except ImportError:
                pass
        try:
            if intro is not None:
                self.intro = intro
            if self.intro:
                self.stdout.write(str(self.intro)+"\n")
            stop = None
            while not stop:
                if self.cmdqueue:
                    line = self.cmdqueue.pop(0)
                else:
                    if self.use_rawinput:
                        line = ''
                        try:
                            line = input(self.prompt)
                        except KeyboardInterrupt:
                            print()
                        except EOFError:
                            line = 'EOF'
                    else:
                        self.stdout.write(self.prompt)
                        self.stdout.flush()
                        line = self.stdin.readline()
                        if not len(line):
                            line = 'EOF'
                        else:
                            line = line.rstrip('\r\n')
                line = self.precmd(line)
                try:
                    stop = self.onecmd(line)
                    stop = self.postcmd(stop, line)
                except Exception as err:
                    print(f"Error: {str(err)}")
            self.postloop()
        finally:
            if self.use_rawinput and self.completekey:
                try:
                    import readline
                    readline.set_completer(self.old_completer)
                except ImportError:
                    pass

    def do_info(self, *args):
        """ Show current session info """
        try:
            print(f"Instance selected: {self.session.instance}")
        except ValueError:
            print("No instance selected")
        else:
            self.session.config.list_configured_instances()
            print(f"Credentials: {self.session.credentials}")
            print(dir(self.session.config))

    def do_connect(self, instance):
        """
        Switch connection to a different instance

        :param instance: Alias for the instance to which you want to connect
        :return:
        """
        log.info(f"Connecting to instance {instance}")
        self.session.instance = instance
        log.info(f"Connected")

    # I want the user to be able to type "exit" instead of having to type "EOF"
    def do_exit(self, line=None):
        return self.do_EOF()

    # What to do when the shell is exited
    def do_EOF(self, line=None):
        return True

    # Override the default behavior with an empty line, which is to rerun the last command. Instead, just do nothing
    def emptyline(self):
        pass

    # Override default output for unknown command to customize the output
    def default(self, line):
        """
        Called on an input line when the command prefix is not recognized.
        Prints an error message and returns.
        """
        self.stdout.write('*** Unknown command: %s\n' % line)

    # Override default help command response to customize the output
    def do_help(self, arg):
        """List available commands with "help" or detailed help with "help cmd"."""
        if arg:
            # XXX check arg syntax
            try:
                func = getattr(self, 'help_' + arg)
            except AttributeError:
                try:
                    doc = getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("%s\n" % str(doc))
                        return
                except AttributeError:
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help_dict = {}
            for name in names:
                if name[:5] == 'help_':
                    help_dict[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    command = name[3:]
                    if command in help_dict:
                        cmds_doc.append(command)
                        del help_dict[command]
                    elif getattr(self, name).__doc__:
                        # cmds_doc.append(command)
                        if command != 'help_dict':
                            cmds_doc.append(command)
                    else:
                        cmds_undoc.append(command)
            self.stdout.write("%s\n" % str(self.doc_leader))
            self.print_topics(self.doc_header,   sorted(cmds_doc),   15, 80)
            self.print_topics(self.misc_header,  list(help_dict.keys()), 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)


# if __name__ == '__main__':
#     Shell().cmdloop()



import sys


def query_yes_no(question, default: (str, bool) = None):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    # Borrowed from here: https://stackoverflow.com/questions/3041986/apt-command-line-interface-like-yes-no-input
    valid_yes = ["yes", "y", "ye"]
    valid_no = ["no", "n"]
    if default is not None:
        if isinstance(default, str):
            default = default.lower().strip()
            assert default in valid_yes + valid_no, f"Invalid input for default value: {default}"
            # Convert to boolean
            default = default in valid_yes

    if default is None:
        prompt = " [y/n] "
    elif default is True:
        prompt = " [Y/n] "
    elif default is False:
        prompt = " [y/N] "
    else:
        raise ValueError(f"Invalid input for default value: {default}")

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower().strip()
        if choice and choice in valid_yes + valid_no:
            return choice in valid_yes
        if not choice and default is not None:
            return default
        else:
            sys.stdout.write("Please respond with one of the following: y/n/yes/no\n\n")



import csv
import json
import os

from tabulate import tabulate, tabulate_formats

SUPPORTED_OUTPUT_TYPES = sorted(["csv", "json", "json_pretty", "table"])
SUPPORTED_TABLE_FORMATS = sorted(tabulate_formats)


def print_fancy_lists(results, output_type="table", table_format=None, ordered_headers=None, output_file=None, sort_order=None, file_only=False):
    """
    Print a list of dicts in a variety of ways, such as json, CSV, or assorted text tables

    :param results: list of dicts with a common schema
    :type results: list

    :param output_type: selection from SUPPORTED_OUTPUT_TYPES (default: table)
    :type output_type: str

    :param table_format: selection from SUPPORTED_TABLE_FORMATS (default: tabulate default)

    :param ordered_headers: optional: customize exact order of columns in the final output
    :type ordered_headers: list

    :param output_file: optional: path to write the output to a file
    :type output_type: str

    :param sort_order: optional: list of column headers to sort the results by before printing
    :type sort_order: list

    """

    def print_json(result_list, pretty=False):
        indent = 2 if pretty else None
        output = json.dumps(result_list, indent=indent)
        if not file_only:
            print(output)
        if output_file:
            with open(output_file, "w+") as _file:
                _file.write(output)

    def print_table(result_list):
        if table_format:
            if table_format not in SUPPORTED_TABLE_FORMATS:
                raise ValueError(f"{table_format} is not a supported table format")

        data = [x.values() for x in result_list]
        headers = ordered_headers
        if not headers:
            headers = result_list[0].keys() if result_list else ['no results']

        output = tabulate(
            tabular_data=data,
            headers=headers,
            tablefmt=table_format or None
        )
        if not file_only:
            print(output)
        if output_file:
            with open(output_file, "w+") as _file:
                _file.write(output)

    def print_csv(result_list):
        headers = result_list[0].keys()

        default_temp_file = '.__temp_csv'
        _output_file = output_file or default_temp_file

        with open(_output_file, 'w') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            for data in result_list:
                writer.writerow(data)

        if not file_only:
            with open(_output_file, 'r') as _file:
                print(_file.read().strip())

        if _output_file == default_temp_file:
            os.remove(_output_file)

    if output_type not in SUPPORTED_OUTPUT_TYPES:
        raise ValueError(f"{output_type} is not a valid output type")

    if sort_order:
        for column in [sort_order[-1 - n] for n in range(len(sort_order))]:
            results = sorted(results, key=lambda e: (e[column]))

    if ordered_headers:
        new_rows = []
        for n in range(len(results)):
            new_entry = {}
            for k in ordered_headers:
                new_entry[k] = results[n][k]
            new_rows.append(new_entry)
        results = new_rows
    if output_type == "json":
        print_json(results, pretty=False)
    elif output_type == "json_pretty":
        print_json(results, pretty=True)
    elif output_type == "table":
        print_table(results)
    elif output_type == "csv":
        print_csv(results)
    else:
        raise ValueError(f"Unsupported output type: {output_type}")



from lhub import LogicHub
from lhub.log import Logger
from lhub.common.dicts_and_lists import to_dict_recursive
import json
from tabulate import tabulate, tabulate_formats
import csv
import os

log = Logger()

# Quick reference for external scripts
TABLE_FORMATS = tabulate_formats


# ToDo Rewrite to make this a part of the LogicHubCLI instance instead of needing the "session" input

class Command:
    __output_type = "json_pretty"
    supported_output_types = ["csv", "json", "json_pretty", "raw", "raw_pretty", "table"]
    verify_ssl = True
    lhub_hidden_fields = ["lhub_page_num", "lhub_id"]

    def __init__(self, session: LogicHub, verify_ssl=True, output_type: str = None, table_format: str = None):
        if output_type:
            self.output_type = output_type
        if isinstance(verify_ssl, bool):
            self.verify_ssl = verify_ssl
        else:
            raise TypeError("verify_ssl requires a boolean value")
        if table_format and table_format not in TABLE_FORMATS:
            raise ValueError(f"Invalid table format: {table_format}")
        self.table_format = table_format
        self.session = session

    @property
    def output_type(self):
        return self.__output_type

    @output_type.setter
    def output_type(self, var):
        if var not in self.supported_output_types:
            raise ValueError(f"\"{var}\" is not a supported output type. Supported types are: {self.supported_output_types}")
        self.__output_type = var

    def __extract_command_results(self, response, fields: list = None, drop: list = None):
        result = response.pop("result")
        warnings = result.pop("warnings", [])
        rows = result['rows']['data']
        drop_fields = self.lhub_hidden_fields
        if drop:
            drop_fields.extend(drop)
        for row_num in range(len(rows)):

            rows[row_num]['fields'] = {k: v for k, v in rows[row_num]['fields'].items() if k not in drop_fields}

            # columns =
            # rows[row_num]['columns'] = {k: v for k, v in rows[row_num]['fields'].items() if k not in drop_fields}
            if fields:
                new_output = {}

                for _field in fields:
                    if _field in rows[row_num]['fields']:
                        new_output[_field] = rows[row_num]['fields'][_field]
                if new_output:
                    rows[row_num]['fields'] = new_output
                else:
                    log.warning(f"None of the provided fields were found in the results. Returning all columns.")
        ordered_headers = []
        if rows:
            ordered_headers = [field['name'] for field in rows[0]['schema']['columns'] if field['name'] not in drop_fields]
        log.debug(f"Non-result response: {json.dumps(response)}")
        return rows, ordered_headers, warnings

    # ToDo Replace most of this with the new helper, common.output.print_fancy_lists
    def print_command_results(self, results, fields: list = None, drop: list = None, output_file: str = None):
        def split_id_from_fields(result_list):
            output = []
            for row in result_list:
                _id = row['id']
                _fields = row['fields']
                log.debug(f"Processing correlation ID: {_id}")
                output.append(_fields)
            return output

        def _print_json(result_list, pretty=False):
            indent = 2 if pretty else None
            output = json.dumps(result_list, indent=indent)
            print(output)
            if output_file:
                with open(output_file, "w+") as _file:
                    _file.write(output)

        def _print_table(result_list):
            if not result_list:
                log.debug("Empty results")
            _keys = result_list[0].keys() if result_list else ['no results']
            kwargs = {
                "tabular_data": [x.values() for x in result_list],
                "headers": _keys
            }
            if self.table_format:
                kwargs["tablefmt"] = self.table_format
            output = tabulate(**kwargs)

            print(output)
            if output_file:
                with open(output_file, "w+") as _file:
                    _file.write(output)

        def _print_raw(result_list, pretty=False):
            indent = 2 if pretty else None
            output = json.dumps(result_list, indent=indent)
            print(output)
            if output_file:
                with open(output_file, "w+") as _file:
                    _file.write(output)

        def _print_csv(result_list):
            headers = result_list[0].keys()

            default_temp_file = '.__temp_csv'
            _output_file = output_file or default_temp_file

            with open(_output_file, 'w') as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=headers)
                writer.writeheader()
                for data in result_list:
                    writer.writerow(data)

            with open(_output_file, 'r') as _file:
                print(_file.read().strip())

            if _output_file == default_temp_file:
                os.remove(_output_file)

        if self.output_type == "raw":
            _print_raw(results, pretty=False)
            return
        elif self.output_type == "raw_pretty":
            _print_raw(results, pretty=True)
            return

        reformatted, ordered_headers, warnings = self.__extract_command_results(results, fields=fields or [], drop=drop or [])

        for _warning in warnings:
            log.warning(f"Warning returned: {_warning}")

        rows = split_id_from_fields(reformatted)
        if ordered_headers:
            new_rows = []
            for n in range(len(rows)):
                new_entry = {}
                for k in ordered_headers:
                    new_entry[k] = rows[n][k]
                new_rows.append(new_entry)
            rows = new_rows
        if self.__output_type == "json":
            _print_json(rows, pretty=False)
        elif self.__output_type == "json_pretty":
            _print_json(rows, pretty=True)
        elif self.__output_type == "table":
            _print_table(rows)
        elif self.__output_type == "csv":
            _print_csv(rows)
        else:
            raise ValueError(f"Unsupported output type: {self.output_type}")

    def run_command(self, command, fix_json=False, fields: list = None, drop: list = None, output_file: str = None, **kwargs):
        fix_json = False if fix_json is False else True
        response = self.session.actions.execute_command(
            command_name=command,
            input_dict=kwargs,
            reformat=False)

        if fix_json:
            response = to_dict_recursive(response)

        self.print_command_results(response, fields=fields, drop=drop, output_file=output_file)



import lhub
import time
import os
from pathlib import Path
from requests import HTTPError
import json
import base64
import re
from .connection_manager import LogicHubConnection
from .common.output import print_fancy_lists


# ToDo NEXT: Follow the same formula from "export_playbooks" to add support for exporting other resource types as well
#  * event types
#  * custom lists
#  * commands
#  * modules
#  * custom integration
#  * baselines
#  * scripts (legacy)
#  * scripts (new, used in runScript)
#  * dashboards & widgets
#  * cases? Probably not, but maybe...


# ToDo This isn't going to scale indefinitely. Need to break these actions apart by feature somehow.
class Actions:

    def __init__(self, session: lhub.LogicHub, config: LogicHubConnection, instance_label):
        self.lhub = session
        self.__config = config
        self.log = self.lhub.api.log
        self.__instance_name = instance_label

    def __set_export_path(self, parent_folder, export_type):
        current_date = time.strftime("%Y-%m-%d")
        _folder_counter = 0
        while True:
            _folder_counter += 1
            _new_export_folder = os.path.join(parent_folder, f"{self.lhub.api.url.server_name}_{export_type}_{current_date}_{_folder_counter}")
            if not os.path.exists(_new_export_folder) or not os.listdir(_new_export_folder):
                parent_folder = _new_export_folder
                path = Path(parent_folder)
                path.mkdir(parents=True, exist_ok=True)
                break
        return parent_folder

    def __save_export_to_disk(self, response, export_folder, resource_id, resource_name, file_info):
        write_mode = "w"
        content_b64 = response["result"]["contentB64"]
        file_type = response["result"]["fileType"]
        file_name = f"{resource_name}.{file_type}"
        # Sanitize the name in case it contains illegal characters for a file name
        file_name = re.sub(r'[^\w\-()\[\] +]', '_', file_name)
        if file_type == "json":
            decoded = base64.b64decode(content_b64).decode("utf-8")
            file_data = json.dumps(json.loads(decoded), indent=4)
        elif file_type == "zip":
            write_mode = "wb"
            file_data = base64.b64decode(content_b64)
        else:
            # Should never happen, but just in case...
            raise lhub.exceptions.LhBaseException(f"\nERROR: Unknown file type. You will need to download manually: {resource_name} ({resource_id})")

        with open(os.path.join(export_folder, file_name), write_mode) as _file:
            _file.write(file_data)
        self.log.info(f"{file_info} - Saved successfully")

    def export_playbooks(self, export_folder, limit=None, return_summary=False):
        export_folder = self.__set_export_path(parent_folder=export_folder, export_type="flows")
        self.log.info(f"Saving files to: {export_folder}")
        flow_ids = self.lhub.actions.playbook_ids
        flow_ids_list = sorted(list(flow_ids.keys()))
        if limit:
            flow_ids_list = flow_ids_list[:limit]
        failed = {}
        for n in range(len(flow_ids_list)):
            _flow_id = flow_ids_list[n]
            _flow_name = flow_ids[_flow_id]
            _file_info = f"{n + 1} of {len(flow_ids_list)}: {_flow_id} ({_flow_name})"
            self.log.info(f"{_file_info} - Downloading...")
            try:
                _response = self.lhub.api.export_playbook(_flow_id)
                self.__save_export_to_disk(response=_response, export_folder=export_folder, resource_id=_flow_id, resource_name=_flow_name, file_info=_file_info)
            except HTTPError:
                warning = f"{_file_info} - Download FAILED"
                _response_message = json.loads(self.lhub.api.last_response_text)
                if not failed.get(_flow_id):
                    failed[_flow_id] = {"name": _flow_name, "errors": []}
                if not _response_message.get("errors"):
                    error = f"unknown failure (status code {self.lhub.api.last_response_status})"
                    new_warning = f"{warning}: {error}"
                    failed[_flow_id]["errors"].append(error)
                    self.log.error(new_warning)
                    with open(os.path.join(export_folder, "_FAILURES.log"), "a+") as _error_file:
                        _error_file.write(new_warning + "\n")
                else:
                    for _error in _response_message.get("errors", []):
                        error = f"{_error.get('errorType')}: {_error['message']}"
                        new_warning = f"{warning}: {error}"
                        failed[_flow_id]["errors"].append(error)
                        self.log.error(new_warning)
                        with open(os.path.join(export_folder, "_FAILURES.log"), "a+") as _error_file:
                            _error_file.write(new_warning + "\n")

        self.log.info("Playbook export complete")
        if return_summary:
            successful = True
            if failed:
                successful = False
            return successful, failed

    def reprocess_batches(self, batch_ids: (list, str, int), sec_between_calls=None):
        if not isinstance(batch_ids, list):
            batch_ids = [batch_ids]
        batch_ids = sorted(list(set(batch_ids)))
        if sec_between_calls:
            sec_between_calls = float(sec_between_calls)
        for n in range(len(batch_ids)):
            batch_id = batch_ids[n]
            if sec_between_calls and n > 0:
                time.sleep(sec_between_calls)
            _ = self.lhub.actions.reprocess_batch(batch_id)
            self.log.info(f'Batch {batch_id} rerun on {self.__instance_name}')

    @staticmethod
    def _reformat_user(user: dict):
        groups = [g['name'] for g in user['groups'] if not g.get("isDeleted", False)]
        user_attributes = {
            "username": user.get("name"),
            "is_admin": user["role"]["value"] == "admin",
            "groups": ', '.join(groups),
            "email": user.get("email"),
            "is_deleted": user.get("isDeleted"),
            "is_enabled": user.get("isEnabled"),
            "auth_type": user.get("authenticationType"),
            "id": user.get("userId"),
        }
        # would it be beneficial to make a class object for these instead of returning a dict?
        return user_attributes

    @staticmethod
    def _reformat_users(users: list):
        return [Actions._reformat_user(user) for user in users]

    def list_users(self, print_output=True, return_results=True, show_hostname=False, sort_order=None, attributes: list = None, hide_inactive=True, **print_kwargs):
        required_columns = ["username"]
        if sort_order is None:
            sort_order = ["connection name", "username"]
        if attributes == "*":
            attributes = []
        attributes = attributes or []

        results = self.lhub.actions.list_users(hide_inactive=hide_inactive, simple_format=True)
        self.log.debug(f"{len([r for r in results if r['is_admin']])} admin users found")

        # Update output to insert
        stock_fields = {"connection name": self.__config.credentials.connection_name}
        if show_hostname:
            stock_fields["hostname"] = self.lhub.api.url.server_name
        results = [
            dict(**stock_fields, **{k: v for k, v in r.items() if k in required_columns or k in attributes or not attributes})
            for r in results
        ]

        for column in [sort_order[-1 - n] for n in range(len(sort_order))]:
            results = sorted(results, key=lambda e: (e[column]))
        if print_output:
            print_fancy_lists(results=results, **print_kwargs)
        if return_results:
            return results

    def list_commands(self, print_output=True, return_results=True, show_hostname=False, sort_order=None, attributes: list = None, **print_kwargs):
        required_columns = ["name"]
        if sort_order is None:
            sort_order = ["connection name", "name"]
        if attributes == "*":
            attributes = []
        attributes = attributes or []

        results = self.lhub.actions.list_commands(simple_format=True)

        # Update output to insert
        stock_fields = {"connection name": self.__config.credentials.connection_name}
        if show_hostname:
            stock_fields["hostname"] = self.lhub.api.url.server_name
        results = [
            dict(**stock_fields, **{k: v for k, v in r.items() if k in required_columns or k in attributes or not attributes})
            for r in results
        ]

        for column in [sort_order[-1 - n] for n in range(len(sort_order))]:
            results = sorted(results, key=lambda e: (e[column]))
        if print_output:
            print_fancy_lists(results=results, **print_kwargs)
        if return_results:
            return results



#!/usr/bin/env python
import argparse

# # create the top-level parser
# parser = argparse.ArgumentParser(prog='PROG')
# parser.add_argument('--foo', action='store_true', help='foo help')
# subparsers = parser.add_subparsers(help='sub-command help')
#
# # create the parser for the "a" command
# parser_a = subparsers.add_parser('a', help='a help')
# parser_a.add_argument('bar', type=int, help='bar help')
#
# # create the parser for the "b" command
# parser_b = subparsers.add_parser('b', help='b help')
# parser_b.add_argument('--baz', choices='XYZ', help='baz help')
#
# # parse some argument lists
# parser.parse_args(['a', '12'])
#
# parser.parse_args(['--foo', 'b', '--baz', 'Z'])
#
# _args = parser.parse_args()

# create the top-level parser
parser = argparse.ArgumentParser(prog='lhub')

# parser.add_argument("x", type=str, help="Some text")

parser.add_argument('--foo', action='store_true', help='foo help')
subparsers = parser.add_subparsers(help='sub-command help')

# create the parser for the "a" command
parser_command = subparsers.add_parser('command', help='command help')
parser_command.add_argument('command', metavar='COMMAND_NAME', type=str, help='command name')

# create the parser for the "b" command
parser_playbook = subparsers.add_parser('playbook', help='playbook help')
parser_playbook.add_argument('playbook', metavar='PLAYBOOK_NAME', type=str, help='playbook name')
# parser_playbook.add_argument('--option', choices='XYZ', help='option help')

# parse some argument lists
# parser.parse_args(['command', '12'])

# parser.parse_args(['--foo', 'b', '--baz', 'Z'])

_args = parser.parse_args()

print(dir(_args))



import argparse
from lhub_cli.common.output import SUPPORTED_OUTPUT_TYPES, SUPPORTED_TABLE_FORMATS


def add_script_output_args(parser: argparse.ArgumentParser, default_output=None):
    if not default_output:
        default_output = "table"
    elif default_output not in SUPPORTED_OUTPUT_TYPES:
        raise ValueError(f"{default_output} is not a supported output type")
    output = parser.add_argument_group('output')

    output.add_argument(
        "-f", "--file",
        type=str,
        default=None,
        help="Also write output to a file")

    output.add_argument(
        "-o", "--output",
        type=str,
        metavar="<OPTION>",
        default=default_output,
        choices=SUPPORTED_OUTPUT_TYPES,
        help=f"Output style (default: {default_output}). Available output types are: {', '.join(SUPPORTED_OUTPUT_TYPES)}")

    # https://github.com/astanin/python-tabulate#table-format
    output.add_argument(
        "-t", "--table_format",
        type=str,
        metavar="<OPTION>",
        default=None,
        choices=SUPPORTED_TABLE_FORMATS,
        help=f"Table format (ignored if output type is not table). Available formats are: {', '.join(SUPPORTED_TABLE_FORMATS)}"
    )



from setuptools import setup, find_packages

VERSION = '0.0.6.4'

setup(
    name="lhub_cli",
    version=VERSION,
    author="Chad Roberts",
    author_email="chad@logichub.com",
    description="LogicHub CLI",
    long_description="A Python package for interacting with LogicHub via shell commands",
    packages=find_packages(),
    install_requires=[
        "lhub == 0.1.6.2",
        "configobj >= 5.0.6",
        "dataclasses_json >= 0.5.6",
        "rsa >= 4.8",
        "requests",
        "tabulate",
        "colorama",
        "progressbar2",  # Only used in sample scripts for now, and imported not as "progressbar2" but just "progressbar"
    ],
    keywords=["python", "lhub", "LogicHub"],

    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Natural Language :: English",
    ]
)



import os
import rsa
import base64
from lhub import exceptions

# https://stuvel.eu/python-rsa-doc/usage.html#generating-keys


class Encryption:
    public_default = ".lhub.pub"
    private_default = ".lhub.pem"

    def __init__(self, key_location, private_key_name=None, pub_file_name=None):
        if not os.path.exists(key_location):
            raise IOError(f"Key location does not exist: {key_location}")
        self.key_location = key_location

        self.private_key_path = os.path.join(
            self.key_location,
            private_key_name.strip() if private_key_name else self.private_default
        )
        self.public_key_path = os.path.join(
            key_location,
            pub_file_name.strip() if pub_file_name else self.public_default
        )

        self._public_key = None
        self.__private_key = None
        self.load_keys()

    def load_keys(self):
        # If no existing keys are found stored at the expected location, generate new ones
        if not os.path.exists(self.private_key_path) and not os.path.exists(self.public_key_path):
            self._public_key, self.__private_key = rsa.newkeys(4096)
            with open(self.public_key_path, "w+") as _key_file:
                _key_file.write(self._public_key.save_pkcs1().decode())
            with open(self.private_key_path, "w+") as _key_file:
                _key_file.write(self.__private_key.save_pkcs1().decode())
            return

        if not os.path.exists(self.private_key_path):
            raise exceptions.EncryptionKeyError(f"Found public key ({self.public_key_path}) but could not find private key ({self.private_key_path})")
        if not os.path.exists(self.public_key_path):
            raise exceptions.EncryptionKeyError(f"Found private key ({self.private_key_path}) but could not find public key ({self.public_key_path})")

        with open(self.public_key_path, mode='rb') as _key_file:
            _public_key_text = _key_file.read()
        with open(self.private_key_path, mode='rb') as _key_file:
            _private_key_text = _key_file.read()
        self._public_key = rsa.PublicKey.load_pkcs1(_public_key_text)
        self.__private_key = rsa.PrivateKey.load_pkcs1(_private_key_text)
        return

    def encrypt_string(self, var_str):
        _var_bytes = var_str.encode()
        _var_encrypted = rsa.encrypt(_var_bytes, self._public_key)
        _var_encoded = base64.b64encode(_var_encrypted)
        return _var_encoded.decode()

    def decrypt_string(self, var_str):
        _var_bytes = var_str.encode()
        _var_decoded = base64.b64decode(_var_bytes)
        _var_decrypted = rsa.decrypt(_var_decoded, self.__private_key)
        return _var_decrypted.decode()
